#
# Copyright 2019, 2022 Hannes Holey
#           2019 Andrea Codrignani
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import yaml
import numpy as np

from hans.problem import Problem
from hans.tools import abort
from hans.material import Material


class Input:

    def __init__(self, inputFile, restartFile=None):
        """
        Reads the yaml input file that defines a problem.

        Parameters
        ----------
        inputFile : str
            filename of the .yaml input file
        restartFile : str
            filename of the .nc data file (default: None)

        """
        self.inputFile = inputFile
        self.restartFile = restartFile

    def getProblem(self):
        """
        Parses the yaml input file, performs sanity checks,
        and returns an instance of the Problem class.

        Returns
        -------
        Problem
            Problem object containing all information about the current job.

        """

        with open(self.inputFile, 'r') as ymlfile:

            inp = yaml.full_load(ymlfile)

            print(f"Reading input from {self.inputFile}...")

            # Mandatory inputs
            options = self.sanitize_options(inp['options'])
            disc = self.sanitize_disc(inp['disc'])
            numerics, disc = self.sanitize_numerics(inp['numerics'], disc)
            geometry = self.sanitize_geometry(inp['geometry'], disc)
            material = self.sanitize_material(inp['material'])
            bc, disc = self.sanitize_BC(inp['BC'], disc, material)

            # Optional inputs
            if "surface" in inp.keys():
                surface = self.sanitize_surface(inp["surface"])
            else:
                surface = None

            if "roughness" in inp.keys():
                roughness = self.sanitize_roughness(inp["roughness"])
            else:
                roughness = None

            if self.restartFile is not None:
                ic = {"type": "restart", "file": self.restartFile}
            elif "IC" in inp.keys():
                ic = self.sanitize_IC(inp["IC"], disc)
            else:
                ic = None

            print("Sanity checks completed. Start simulation!")
            print(60 * "-")

        thisProblem = Problem(options,
                              disc,
                              bc,
                              geometry,
                              numerics,
                              material,
                              surface,
                              ic,
                              roughness)

        return thisProblem

    def sanitize_disc(self, disc):
        """Sanity check for discretization settings

        Parameters
        ----------
        disc : dict
            Discretization input dict

        Returns
        -------
        dict
            Sanitized discretization input dict
        """

        print("Checking discretization... ")

        try:
            disc["Nx"] = int(disc['Nx'])
            assert disc["Nx"] > 0
        except KeyError:
            print("***Number of grid cells Nx not specified. Abort.")
            abort()
        except AssertionError:
            print("***Number of grid cells Nx must be larger than zero. Abort")
            abort()

        try:
            disc["Ny"] = int(disc['Ny'])
            assert disc["Ny"] > 0
        except KeyError:
            print("***Number of grid cells 'Ny' not specified. Abort.")
            abort()
        except AssertionError:
            print("***Number of grid cells 'Ny' must be larger than zero. Abort")
            abort()

        try:
            disc["Lx"] = float(disc["Lx"])
        except KeyError:
            try:
                disc["dx"] = float(disc["dx"])
            except KeyError:
                print("At least two of 'Nx' 'Lx', 'dx' must be given. Abort.")
                abort()
            else:
                disc["Lx"] = disc["dx"] * disc["Nx"]
        else:
            disc["dx"] = disc["Lx"] / disc["Nx"]

        try:
            disc["Ly"] = float(disc["Ly"])
        except KeyError:
            try:
                disc["dy"] = float(disc["dy"])
            except KeyError:
                print("At least two of 'Ny' 'Ly', 'dy' must be given. Abort.")
                abort()
            else:
                disc["Ly"] = disc["dy"] * disc["Ny"]
        else:
            disc["dy"] = disc["Ly"] / disc["Ny"]

        return disc

    def sanitize_options(self, options):
        """Sanitize output settings


        Parameters
        ----------
        options : dict
            Output settings

        Returns
        -------
        dict
            Sanitized output settings
        """
        print("Checking I/O options... ")

        try:
            writeInterval = int(options["writeInterval"])
            assert writeInterval > 0
        except KeyError:
            print("***Output interval not given, fallback to 1000")
            options["writeInterval"] = 1000
        except AssertionError:
            try:
                assert writeInterval != 0
            except AssertionError:
                print("***Output interval is zero. fallback to 1000")
                options["writeInterval"] = 1000
            else:
                print("***Output interval is negative. Converting to positive value.")
                writeInterval *= -1
                options["writeInterval"] = writeInterval

        return options

    def sanitize_geometry(self, geometry, disc):
        """Sanitize geometry

        Parameters
        ----------
        geometry : dict
            Geometry settings
        disc : dict
            Discretization settings

        Returns
        -------
        dict
            Sanitized geometry settings
        """

        print("Checking geometry... ")

        if geometry["type"] in ["journal", "journal_x", "journal_y"]:
            geometry["CR"] = float(geometry["CR"])
            geometry["eps"] = float(geometry["eps"])
        elif geometry["type"] == "parabolic":
            geometry["hmin"] = float(geometry['hmin'])
            geometry["hmax"] = float(geometry['hmax'])
        elif geometry["type"] == "twin_parabolic":
            geometry["hmin"] = float(geometry['hmin'])
            geometry["hmax"] = float(geometry['hmax'])
        elif geometry["type"] in ["inclined", "inclined_x", "inclined_y"]:
            geometry["h1"] = float(geometry['h1'])
            geometry["h2"] = float(geometry['h2'])
        elif geometry["type"] == "inclined_pocket":
            geometry["h1"] = float(geometry['h1'])
            geometry["h2"] = float(geometry['h2'])
            geometry["hp"] = float(geometry['hp'])
            geometry["c"] = float(geometry['c'])
            geometry["l"] = float(geometry['l'])
            geometry["w"] = float(geometry['w'])

            try:
                geometry["t"] = float(geometry["t"])
            except KeyError:
                geometry["t"] = 0.

            try:
                assert 2 * geometry["t"] + geometry["l"] + geometry["c"] <= disc["Lx"]
                assert geometry["w"] <= disc["Ly"]
            except AssertionError:
                print("Size of the pocket is larger than the domain. Abort.")
                abort()

        elif geometry["type"] in ["half_sine", "half_sine_squared"]:
            geometry["h0"] = float(geometry['h0'])
            geometry["amp"] = float(geometry['amp'])
            geometry["num"] = float(geometry['num'])

        else:
            print(f"'{geometry['type']}' geometry not implemented")
            abort()

        return geometry

    def sanitize_numerics(self, numerics, disc):
        """Sanitize numerics settings

        Parameters
        ----------
        numerics : dict
            Numerics input settings
        disc : dict
            Discretization settings

        Returns
        ------- 
        dict
            Sanitized numerics settings
        dict
            Modified discretization settings
        """

        print("Checking numerics options... ")

        try:
            numerics["integrator"] = numerics["integrator"]
            assert numerics["integrator"] in ["MC", "MC_bf", "MC_fb", "MC_alt", "LW", "RK3"]
        except KeyError:
            print("***Integrator not specified. Use default (MacCormack).")
            numerics["integrator"] = "MC"
        except AssertionError:
            print(f'***Unknown integrator \'{numerics["integrator"]}\'. Abort.')
            abort()

        if numerics["integrator"].startswith("MC"):
            try:
                numerics["fluxLim"] = float(numerics["fluxLim"])
            except KeyError:
                pass

        try:
            numerics["stokes"] = int(numerics["stokes"])
        except KeyError:
            print("***Boolean parameter 'stokes' not given. Use default (True).")
            numerics["stokes"] = 1

        try:
            numerics["adaptive"] = int(numerics["adaptive"])
        except KeyError:
            print("***Boolean parameter 'adaptive' not given. Use default (False).")
            numerics["adaptive"] = 0

        if numerics["adaptive"] == 1:
            try:
                numerics["C"] = float(numerics["C"])
            except KeyError:
                print("***CFL number not given. Use default (0.5).")
                numerics["C"] = 0.5

        try:
            numerics["dt"] = float(numerics["dt"])
        except KeyError:
            print("***Timestep not given. Use default (1e-10).")
            numerics["dt"] = 1e-10

        stopping_criteria = 0

        try:
            numerics["tol"] = float(numerics["tol"])
            stopping_criteria += 1
        except KeyError:
            pass

        try:
            numerics["maxT"] = float(numerics["maxT"])
            stopping_criteria += 1
        except KeyError:
            pass

        try:
            numerics["maxIt"] = int(numerics["maxIt"])
            stopping_criteria += 1
        except KeyError:
            pass

        if stopping_criteria == 0:
            print("***No stopping criterion given. Abort.")
            abort()

        if numerics["integrator"] == "RK3":
            disc["nghost"] = 2
        else:
            disc["nghost"] = 1

        return numerics, disc

    def sanitize_material(self, material):
        """Sanitize material settings

        Parameters
        ----------
        material : dict
            Material (fluid) properties

        Returns
        -------
        dict
            Sanitized material properties
        """

        print("Checking material options... ")

        if material["EOS"] == "DH":
            material["rho0"] = float(material["rho0"])
            material["P0"] = float(material["P0"])
            material["C1"] = float(material["C1"])
            material["C2"] = float(material["C2"])
        elif material["EOS"] == "PL":
            material["rho0"] = float(material["rho0"])
            material["P0"] = float(material["P0"])
            material["alpha"] = float(material['alpha'])
        elif material["EOS"] == "vdW":
            material["M"] = float(material['M'])
            material["T"] = float(material['T0'])
            material["a"] = float(material['a'])
            material["b"] = float(material['b'])
        elif material["EOS"] == "Tait":
            material["rho0"] = float(material["rho0"])
            material["P0"] = float(material["P0"])
            material["K"] = float(material['K'])
            material["n"] = float(material['n'])
        elif material["EOS"] == "cubic":
            material["a"] = float(material['a'])
            material["b"] = float(material['b'])
            material["c"] = float(material['c'])
            material["d"] = float(material['d'])
        elif material["EOS"].startswith("Bayada"):
            material["cl"] = float(material["cl"])
            material["cv"] = float(material["cv"])
            material["rhol"] = float(material["rhol"])
            material["rhov"] = float(material["rhov"])
            material["shear"] = float(material["shear"])
            material["shearv"] = float(material["shearv"])
            material["rhov"] = float(material["rhov"])

        material["shear"] = float(material["shear"])
        material["bulk"] = float(material["bulk"])

        if "Pcav" in material.keys():
            material["Pcav"] = float(material["Pcav"])

        if "piezo" in material.keys():
            if material["piezo"] == "Barus":
                material["aB"] = float(material["aB"])
            elif material["piezo"] == "Vogel":
                material["rho0"] = float(material['rho0'])
                material["g"] = float(material["g"])
                material["mu_inf"] = float(material["mu_inf"])
                material["phi_inf"] = float(material["phi_inf"])
                material["BF"] = float(material["BF"])

        if "thinning" in material.keys():
            if material["thinning"] == "Eyring":
                material["tau0"] = float(material["tau0"])
            elif material["thinning"] == "Carreau":
                material["relax"] = float(material["relax"])
                material["a"] = float(material["a"])
                material["N"] = float(material["N"])
            elif material["thinning"] == "PL":
                material["shear"] = float(material["shear"])
                material["n"] = float(material["n"])

        if "PLindex" in material.keys():
            material["PLindex"] = float(material["PLindex"])
            if "PLmethod" in material.keys():
                material["PLmethod"] = material["PLmethod"]
            else:
                material["PLmethod"] = "exact"

        return material

    def sanitize_BC(self, bc, disc, material):
        """Sanity check for boundary condition input.

        Parameters
        ----------
        bc : dict
            Boundary condition parameters read from yaml input file.
        disc : dict
            Discretization parameters.
        material : dict
            Material parameters.

        Returns
        -------
        dict
            Boundary condition parameters.
        dict
            Modified discretization parameters.

        """
        print("Checking boundary conditions... ")

        bc["x0"] = np.array(list(bc["x0"]))
        bc["x1"] = np.array(list(bc["x1"]))
        bc["y0"] = np.array(list(bc["y0"]))
        bc["y1"] = np.array(list(bc["y1"]))

        assert len(bc["x0"]) == 3
        assert len(bc["x1"]) == 3
        assert len(bc["y0"]) == 3
        assert len(bc["y1"]) == 3

        if "P" in bc["x0"] and "P" in bc["x1"]:
            disc["pX"] = 1
        else:
            disc["pX"] = 0

        if "P" in bc["y0"] and "P" in bc["y1"]:
            disc["pY"] = 1
        else:
            disc["pY"] = 0

        if "D" in bc["x0"]:
            if "px0" in bc.keys():
                px0 = float(bc["px0"])
                bc["rhox0"] = Material(material).eos_density(px0)
            elif "rhox0" in bc.keys():
                rhox0 = float(bc["rhox0"])
                bc["rhox0"] = rhox0
            else:
                bc["rhox0"] = material["rho0"]

        if "D" in bc["x1"]:
            if "px1" in bc.keys():
                px1 = float(bc["px1"])
                bc["rhox1"] = Material(material).eos_density(px1)
            elif "rhox1" in bc.keys():
                rhox1 = float(bc["rhox1"])
                bc["rhox1"] = rhox1
            else:
                bc["rhox1"] = material["rho0"]

        if "D" in bc["y0"]:
            if "py0" in bc.keys():
                py0 = float(bc["py0"])
                bc["rhoy0"] = Material(material).eos_density(py0)
            elif "rhoy0" in bc.keys():
                rhoy0 = float(bc["rhoy0"])
                bc["rhoy0"] = rhoy0
            else:
                bc["rhoy0"] = material["rho0"]

        if "D" in bc["y1"]:
            if "py1" in bc.keys():
                py1 = float(bc["py1"])
                bc["rhoy1"] = Material(material).eos_density(py1)
            elif "rhoy1" in bc.keys():
                rhoy1 = float(bc["rhoy1"])
                bc["rhoy1"] = rhoy1
            else:
                bc["rhoy1"] = material["rho0"]

        assert np.all((bc["x0"] == "P") == (bc["x1"] == "P")), "Inconsistent boundary conditions (x)"
        assert np.all((bc["y0"] == "P") == (bc["y1"] == "P")), "Inconsistent boundary conditions (y)"

        return bc, disc

    def sanitize_surface(self, surface):
        """Sanitize surface parameters

        Parameters
        ----------
        surface : dict
            Input surface parameters

        Returns
        -------
        dict
            Sanitized surface parameters
        """
        print("Checking surface parameters... ")

        if "lslip" in surface.keys():
            surface["lslip"] = float(surface["lslip"])
        else:
            surface["lslip"] = 0.

        if surface["type"] in ["stripes", "stripes_x", "stripes_y"]:
            try:
                surface["num"] = int(surface["num"])
            except KeyError:
                surface["num"] = 1
            try:
                surface["sign"] = int(surface["sign"])
            except KeyError:
                surface["sign"] = -1

        return surface

    def check_roughness(self, roughness):
        """Sanitize roughness settings

        Parameters
        ----------
        roughness : dict
            Roughness settings

        Returns
        -------
        dict
            Sanitized roughness settings
        """

        print("Checking roughness parameters... ")

        if "file" not in roughness.keys():
            try:
                roughness["seed"] = int(roughness["seed"])
            except KeyError:
                roughness["seed"] = None
            try:
                roughness["Hurst"] = float(roughness["Hurst"])
            except KeyError:
                "Hurst exponent not given. Use default (0.8)."
                roughness["Hurst"] = 0.8
            try:
                roughness["rolloff"] = float(roughness["rolloff"])
            except KeyError:
                "Rollof not given. Use default (1.0)"
                roughness["rolloff"] = 1.0
            try:
                roughness["rmsHeight"] = float(roughness["rmsHeight"])
            except KeyError:
                roughness["rmsHeight"] = None
            try:
                roughness["rmsSlope"] = float(roughness["rmsSlope"])
            except KeyError:
                roughness["rmsSlope"] = None
            try:
                roughness["shortCutoff"] = float(roughness["shortCutoff"])
            except KeyError:
                roughness["shortCutoff"] = None
            try:
                roughness["longCutoff"] = float(roughness["longCutoff"])
            except KeyError:
                roughness["longCutoff"] = None

            if roughness["rmsHeight"] is None and roughness["rmsSlope"] is None:
                print('Neither rms height nor rms slope is defined! Abort.')
                abort()

        return roughness

    def sanitize_IC(self, ic, disc):
        """Sanitize initial conditions

        Parameters
        ----------
        ic : dict
            Initial conditions settings
        disc : dict
            Discretization settings

        Returns
        -------
        dict
            Sanitized initial conditions setting
        """
        print("Checking initial conditions... ")

        if ic["type"] != "restart":
            if ic["type"] == "perturbation":
                ic["factor"] = float(ic["factor"])
            elif ic["type"] in ["longitudinal_wave", "shear_wave"]:
                ic["amp"] = float(ic["amp"])
                if "nwave" in ic.keys():
                    ic["nwave"] = int(ic["nwave"])
                else:
                    ic["nwave"] = 1
            elif ic["type"] == "random":
                try:
                    ic["stdDens"] = float(ic["stdDens"])
                except KeyError:
                    ic["stdDens"] = 0.0
                try:
                    ic["stdFlux"] = float(ic["stdFlux"])
                except KeyError:
                    ic["stdFlux"] = 0.0
            elif ic["type"] == "gauss1D":
                try:
                    ic["mean"] = float(ic["mean"])
                except KeyError:
                    ic["mean"] = float(disc["Lx"]) / 2
                try:
                    ic["stdev"] = float(ic["stdev"])
                except KeyError:
                    ic["stdev"] = float(disc["Lx"]) / 40
                try:
                    ic["cutoff"] = float(ic["cutoff"])
                except KeyError:
                    ic["cutoff"] = 12 * ic['stdev']

        return ic



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np
from mpi4py import MPI

from hans.field import VectorField
from hans.stress import SymStressField2D, SymStressField3D
from hans.geometry import GapHeight, SlipLength
from hans.material import Material


class ConservedField(VectorField):

    def __init__(self, disc, bc, geometry, material, numerics, surface, roughness,
                 q_init=None, t_init=None):
        """
        This class contains the field of conserved variable densities (rho, jx, jy),
        and the methods to update them. Derived from VectorField.

        Parameters
        ----------
        disc : dict
            Discretization parameters.
        bc : dict
            Boundary condition parameters.
        geometry : dict
            Geometry parameters.
        material : dict
            Material parameters.
        numerics : dict
            numerics parameters
        surface : dict
            surface parameters
        roughness : dict
            roughness parameters
        q_init : np.array
            Inital field in case of a restart (the default is None)
        t_init : tuple
            Initial time and time step in case of a restart (the default is None).

        """

        super().__init__(disc)

        self.disc = disc
        self.bc = bc
        self.material = material
        self.numerics = numerics
        self.roughness = roughness

        # initialize gap height and slip length field
        self.height = GapHeight(disc, geometry, roughness)
        self.slip_length = SlipLength(disc, surface)

        # intialize field and time
        if q_init is not None:
            self.field[:, 1:-1, 1:-1] = q_init[:, self.without_ghost[0], self.without_ghost[1]]
            self.fill_ghost_buffer()
            self.time = t_init[0]
            self.dt = t_init[1]
        else:
            self.field[0] = material['rho0']
            self.time = 0.
            self.dt = numerics["dt"]

        self.eps = np.nan

        self.viscous_stress = SymStressField2D(disc, geometry, material, surface)
        self.upper_stress = SymStressField3D(disc, geometry, material, surface)
        self.lower_stress = SymStressField3D(disc, geometry, material, surface)

    @property
    def mass(self):
        area = self.disc["dx"] * self.disc["dy"]
        local_mass = np.sum(self.inner[0] * self.height.inner[0] * area)
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_mass, recvbuf, op=MPI.SUM)

        return recvbuf[0]

    @property
    def vSound(self):
        local_vSound = Material(self.material).eos_sound_speed(self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_vSound, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def vmax(self):
        local_vmax = np.amax(np.sqrt(self.inner[1]**2 + self.inner[2]**2) / self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_vmax, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def vx_max(self):
        local_vx_max = np.amax(self.inner[1] / self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_vx_max, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def vy_max(self):
        local_vy_max = np.amax(self.inner[2] / self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_vy_max, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def ekin(self):
        area = self.disc["dx"] * self.disc["dy"]
        local_ekin = np.sum((self.inner[1]**2 + self.inner[2]**2) / self.inner[0] * area)
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_ekin, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def isnan(self):
        local_nans = np.sum(np.isnan(self.inner))
        recvbuf = np.empty(1, dtype=int)
        self.comm.Allreduce(local_nans, recvbuf, op=MPI.SUM)

        return recvbuf[0]

    def update(self, i):
        """
        Wrapper function for the explicit time update of the solution field.

        Parameters
        ----------
        i : int
            time step

        """

        integrator = self.numerics["integrator"]

        # MacCormack forward backward
        if integrator == "MC" or integrator == "MC_fb":
            try:
                self.mac_cormack(0)
            except NotImplementedError:
                print("MacCormack scheme not implemented. Exit!")
                quit()

        # Mac Cormack backward forward
        elif integrator == "MC_bf":
            try:
                self.mac_cormack(1)
            except NotImplementedError:
                print("MacCormack scheme not implemented. Exit!")
                quit()

        # MacCormack alternating
        elif integrator == "MC_alt":
            try:
                self.mac_cormack(i % 2)
            except NotImplementedError:
                print("MacCormack scheme not implemented. Exit!")
                quit()

        # Richtmyer two-step Lax-Wendroff
        elif integrator == "LW":
            try:
                self.richtmyer()
            except NotImplementedError:
                print("Lax-Wendroff scheme not implemented. Exit!")
                quit()

        # 3rd order Runge-Kutta
        elif integrator == "RK3":
            try:
                self.runge_kutta3()
            except NotImplementedError:
                print("Runge-Kutta 3 scheme not implemented. Exit!")
                quit()

    def mac_cormack(self, switch):
        """
        Explicit time update using MacCormack's predictor corrector scheme with source term.
        Field is updated in place.

        Parameters
        ----------
        switch : int
            Determine if predictor step is forwards (0) or backwards (1).

        """

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        if switch == 0:
            directions = [-1, 1]
        elif switch == 1:
            directions = [1, -1]

        q0 = self.field.copy()

        for dir in directions:

            fX, fY = self.predictor_corrector(self.field, self.height.field, dir)
            src = self.get_source(self.field, self.height.field, self.slip_length.field)

            self.field = self.field - self.dt * (fX / dx + fY / dy - src)
            self.fill_ghost_buffer()

        self.field = 0.5 * (self.field + q0)

        if "fluxLim" in self.numerics.keys():
            self.field += self.numerics["fluxLim"] * self.TVD_MC_correction(q0)
            self.fill_ghost_buffer()

        self.post_integrate(q0)

    def TVD_MC_correction(self, q):
        """
        Compute the correction for the TVD-MacCormack scheme (Davis, 1984).

        Parameters
        ----------
        q : np.array
            Copy of the solution field from the previous step

        Returns
        ----------
        np.array
            TVD correction term
        """

        q_diffx_pos = np.roll(q, -1, axis=1) - q
        denom_x_pos = np.sum(q_diffx_pos**2, axis=0)
        q_diffx_neg = q - np.roll(q, 1, axis=1)
        denom_x_neg = np.sum(q_diffx_neg**2, axis=0)

        nonzero_x_pos = denom_x_pos != 0
        nonzero_x_neg = denom_x_neg != 0

        rx_pos = np.ones_like(denom_x_pos)
        rx_pos[nonzero_x_pos] = np.sum(q_diffx_neg * q_diffx_pos, axis=0)[nonzero_x_pos] / denom_x_pos[nonzero_x_pos]

        rx_neg = np.ones_like(denom_x_neg)
        rx_neg[nonzero_x_neg] = np.sum(q_diffx_neg * q_diffx_pos, axis=0)[nonzero_x_neg] / denom_x_neg[nonzero_x_neg]

        q_diffy_pos = np.roll(q, -1, axis=2) - q
        denom_y_pos = np.sum(q_diffx_pos**2, axis=0)
        q_diffy_neg = q - np.roll(q, 1, axis=2)
        denom_y_neg = np.sum(q_diffx_pos**2, axis=0)

        nonzero_y_pos = denom_y_pos != 0
        nonzero_y_neg = denom_y_neg != 0

        ry_pos = np.ones_like(denom_y_pos)
        ry_neg = np.ones_like(denom_y_pos)

        ry_pos[nonzero_y_pos] = np.sum(q_diffy_neg * q_diffy_pos, axis=0)[nonzero_y_pos] / denom_y_pos[nonzero_y_pos]
        ry_neg[nonzero_y_neg] = np.sum(q_diffy_neg * q_diffy_pos, axis=0)[nonzero_y_neg] / denom_y_neg[nonzero_y_neg]

        G_rx_pos = self.flux_limiter_function(rx_pos)
        G_rx_pos_W = self.flux_limiter_function(np.roll(rx_pos, 1, axis=0))

        G_rx_neg = self.flux_limiter_function(rx_neg)
        G_rx_neg_E = self.flux_limiter_function(np.roll(rx_neg, -1, axis=0))

        G_ry_pos = self.flux_limiter_function(ry_pos)
        G_ry_pos_S = self.flux_limiter_function(np.roll(ry_pos, 1, axis=1))

        G_ry_neg = self.flux_limiter_function(rx_neg)
        G_ry_neg_N = self.flux_limiter_function(np.roll(ry_neg, -1, axis=1))

        return (G_rx_pos + G_rx_neg_E) * q_diffx_pos \
            - (G_rx_pos_W + G_rx_neg) * q_diffx_neg \
            + (G_ry_pos + G_ry_neg_N) * q_diffy_pos \
            - (G_ry_pos_S + G_ry_neg) * q_diffy_neg

    def flux_limiter_function(self, r):
        """Flux limiter function of the TCD-MacCormack scheme (Davis, 1984)

        Parameters
        ----------
        r : np.array
            Scalar field, quantifying the relative cell difference of neighboring cell

        Returns
        ----------
        np.array
            Values of the flux limiter function
        """

        cfl = self.dt * (self.vmax + self.vSound) / min(self.disc["dx"], self.disc["dy"])

        if cfl <= 0.5:
            C = cfl * (1 - cfl)
        else:
            C = 0.25

        phi = np.maximum(np.zeros_like(r), np.minimum(2*r, 1))

        return 0.5 * C * (1 - phi)

    def richtmyer(self):
        """
        Explicit time update using Richtmyer's two-step scheme with source term.
        Field is updated in place.

        """

        q0 = self.field.copy()

        # Step 1
        self.lax_step()
        self.fill_ghost_buffer()

        # Step 2
        self.leap_frog(q0)
        self.fill_ghost_buffer()

        self.post_integrate(q0)

    def runge_kutta3(self):

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        q0 = self.field.copy()

        for stage in np.arange(1, 4):

            dX, dY = self.diffusiveCD()
            fX, fY = self.hyperbolicTVD()
            src = self.get_source(self.field, self.height.field, self.slip_length.field)

            tmp = self.field - self.dt * (fX / dx + fY / dy + dX / (2 * dx) + dY / (2 * dy) - src)

            if stage == 1:
                self.field = tmp
                self.fill_ghost_buffer()
            elif stage == 2:
                self.field = 3 / 4 * q0 + 1 / 4 * tmp
                self.fill_ghost_buffer()
            elif stage == 3:
                self.field = 1 / 3 * q0 + 2 / 3 * tmp
                self.fill_ghost_buffer()

        self.post_integrate(q0)

    def post_integrate(self, q0):
        """
        Compute relative change of solution after time update.
        Update time, and, if adaptive time-stepping is used, also the time step.

        Parameters
        ----------
        q0 : np.array
            Copy of the solution field from the previous step

        """

        ng = self.disc["nghost"]

        self.time += self.dt

        dx = np.array([self.disc["dx"], self.disc["dy"]])
        vmax = np.array([self.vx_max, self.vy_max]) + self.vSound
        dt_crit = dx / vmax

        if bool(self.numerics["adaptive"]):
            CFL = self.numerics["C"]
            self.dt = CFL * np.amin(dt_crit)
        else:
            CFL = self.dt / np.amax(dt_crit)

        diff = np.empty(1)
        denom = np.empty(1)

        local_diff = np.sum((self.inner[0] - q0[0, ng:-ng, ng:-ng])**2)
        local_denom = np.sum(q0[0, ng:-ng, ng:-ng]**2)

        self.comm.Allreduce(local_diff, diff, op=MPI.SUM)
        self.comm.Allreduce(local_denom, denom, op=MPI.SUM)

        self.eps = np.sqrt(diff[0] / denom[0]) / CFL

    def fill_ghost_buffer(self):
        """
        Communicate results from adjacent domains to ghost buffers.
        Fill ghost cells at the boundary according to the applied BCs.

        """

        x0 = self.bc["x0"]
        x1 = self.bc["x1"]
        y0 = self.bc["y0"]
        y1 = self.bc["y1"]

        (ls, ld), (rs, rd), (bs, bd), (ts, td) = self.get_neighbors()

        ng = self.disc["nghost"]
        ngt = 2 * ng

        # Send to left, receive from right
        recvbuf = np.ascontiguousarray(self.field[:, -ng:, :])
        self.comm.Sendrecv(np.ascontiguousarray(self.field[:, ng:ngt, :]), ld, recvbuf=recvbuf, source=ls)
        # fill ghost buffer, adjust for non-periodic BCs
        if ls >= 0:
            self.field[:, -ng:, :] = recvbuf
        else:
            if np.any(x1 == "D"):
                self.field[x1 == "D", -1, :] = 2. * self.bc["rhox1"] - self.field[x1 == "D", -2, :]
            if np.any(x1 == "N"):
                self.field[x1 == "N", -1, :] = self.field[x1 == "N", -2, :]

        # Send to right, receive from left
        recvbuf = np.ascontiguousarray(self.field[:, :ng, :])
        self.comm.Sendrecv(np.ascontiguousarray(self.field[:, -ngt:-ng, :]), rd, recvbuf=recvbuf, source=rs)
        # fill ghost buffer, adjust for non-periodic BCs
        if rs >= 0:
            self.field[:, :ng, :] = recvbuf
        else:
            if np.any(x0 == "D"):
                self.field[x0 == "D", 0, :] = 2. * self.bc["rhox0"] - self.field[x0 == "D", 1, :]
            if np.any(x0 == "N"):
                self.field[x0 == "N", 0, :] = self.field[x0 == "N", 1, :]

        # Send to bottom, receive from top
        recvbuf = np.ascontiguousarray(self.field[:, :, -ng:])
        self.comm.Sendrecv(np.ascontiguousarray(self.field[:, :, ng:ngt]), bd, recvbuf=recvbuf, source=bs)
        # fill ghost buffer, adjust for non-periodic BCs
        if bs >= 0:
            self.field[:, :, -ng:] = recvbuf
        else:
            if np.any(y1 == "D"):
                self.field[y1 == "D", :, -1] = 2. * self.bc["rhoy1"] - self.field[y1 == "D", :, -2]
            if np.any(y1 == "N"):
                self.field[y1 == "N", :, -1] = self.field[y1 == "N", :, -2]

        # Send to top, receive from bottom
        recvbuf = np.ascontiguousarray(self.field[:, :, :ng])
        self.comm.Sendrecv(np.ascontiguousarray(self.field[:, :, -ngt:-ng]), td, recvbuf=recvbuf, source=ts)
        # fill ghost buffer, adjust for non-periodic BCs
        if ts >= 0:
            self.field[:, :, :ng] = recvbuf
        else:
            if np.any(y0 == "D"):
                self.field[y0 == "D", :, 0] = 2. * self.bc["rhoy0"] - self.field[y0 == "D", :, 1]
            if np.any(y0 == "N"):
                self.field[y0 == "N", :, 0] = self.field[y0 == "N", :, 1]

    def get_source(self, q, h, Ls):
        """
        Compute the source term from the unknown and gap height.

        Parameters
        ----------
        q : np.array
            Solution field.
        h : np.array
            Gap height (and gradients).
        Ls : np.array
            Slip length (lower surface)

        Returns
        -------
        np.array
            Source term

        """

        self.viscous_stress.set(q, h, Ls)
        self.upper_stress.set(q, h, Ls, "top")
        self.lower_stress.set(q, h, Ls, "bottom")

        stress = self.viscous_stress.field

        out = np.zeros_like(q)

        # origin bottom, U_top = 0, U_bottom = U
        out[0] = (-q[1] * h[1] - q[2] * h[2]) / h[0]

        if bool(self.numerics["stokes"]):
            out[1] = ((stress[0] - self.upper_stress.field[0]) * h[1] +
                      (stress[2] - self.upper_stress.field[5]) * h[2] +
                      self.upper_stress.field[4] - self.lower_stress.field[4]) / h[0]

            out[2] = ((stress[2] - self.upper_stress.field[5]) * h[1] +
                      (stress[1] - self.upper_stress.field[1]) * h[2] +
                      self.upper_stress.field[3] - self.lower_stress.field[3]) / h[0]
        else:
            out[1] = ((q[1] * q[1] / q[0] + stress[0] - self.upper_stress.field[0]) * h[1] +
                      (q[1] * q[2] / q[0] + stress[2] - self.upper_stress.field[5]) * h[2] +
                      self.upper_stress.field[4] - self.lower_stress.field[4]) / h[0]

            out[2] = ((q[2] * q[1] / q[0] + stress[2] - self.upper_stress.field[5]) * h[1] +
                      (q[2] * q[2] / q[0] + stress[1] - self.upper_stress.field[1]) * h[2] +
                      self.upper_stress.field[3] - self.lower_stress.field[3]) / h[0]

        return out

    def predictor_corrector(self, q, h, dir):
        """
        Compute predictor/corrector fluxes fro MacCormack's method.

        Parameters
        ----------
        q : np.array
            Solution field.
        h : np.array
            Gap height (and gradients).
        dir : int
            Direction of the substep (-1: forwards / 1: backwards differencing)

        Returns
        -------
        tuple
            Total numerical flux in x- and y-directions respectively.

        """

        Fx = self.hyperbolicFlux(q, 1) - self.diffusiveFlux(q, h, self.slip_length.field, 1)
        Fy = self.hyperbolicFlux(q, 2) - self.diffusiveFlux(q, h, self.slip_length.field, 2)

        flux_x = -dir * (np.roll(Fx, dir, axis=1) - Fx)
        flux_y = -dir * (np.roll(Fy, dir, axis=2) - Fy)

        return flux_x, flux_y

    def hyperbolicFlux(self, q, ax):
        """Compute hyperbolic flux.

        Parameters
        ----------
        q : np.array
            Solution field.
        ax : int
            Axis parameter, 1 == x and 2 == y

        Returns
        -------
        np.array
            Hyperbolic flux field.

        """

        p = Material(self.material).eos_pressure(q[0])
        inertialess = bool(self.numerics["stokes"])

        F = np.zeros_like(q)
        if ax == 1:
            if inertialess:
                F[0] = q[1]
                F[1] = p
            else:
                F[0] = q[1]
                F[1] = q[1] * q[1] / q[0] + p
                F[2] = q[2] * q[1] / q[0]

        elif ax == 2:
            if inertialess:
                F[0] = q[2]
                F[2] = p
            else:
                F[0] = q[2]
                F[1] = q[1] * q[2] / q[0]
                F[2] = q[2] * q[2] / q[0] + p

        return F

    def diffusiveFlux(self, q, h, Ls, ax):
        """Compute diffusive flux.

        Parameters
        ----------
        q : np.array
            Solution field.
        h : np.array
            Gap height (and gradients).
        Ls : np.array
            Slip length (lower surface)
        ax : int
            Axis parameter, 1 == x and 2 == y

        Returns
        -------
        np.array
            Diffusive flux field

        """

        self.viscous_stress.set(q, h, Ls)

        D = np.zeros_like(q)
        if ax == 1:
            D[1] = self.viscous_stress.field[0]
            D[2] = self.viscous_stress.field[2]

        elif ax == 2:
            D[1] = self.viscous_stress.field[2]
            D[2] = self.viscous_stress.field[1]

        return D

    def lax_step(self):

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        FE = self.hyperbolicFlux(self.edgeE, 1) - self.diffusiveFlux(self.edgeE, self.height.edgeE, self.slip_length.field, 1)
        FN = self.hyperbolicFlux(self.edgeN, 2) - self.diffusiveFlux(self.edgeN, self.height.edgeN, self.slip_length.field, 2)
        FW = self.hyperbolicFlux(self.edgeW, 1) - self.diffusiveFlux(self.edgeW, self.height.edgeW, self.slip_length.field, 1)
        FS = self.hyperbolicFlux(self.edgeS, 2) - self.diffusiveFlux(self.edgeS, self.height.edgeS, self.slip_length.field, 2)
        src = self.get_source(self.field, self.height.field, self.slip_length.field)

        self.field = self.field - self.dt / 2. * ((FE - FW) / dx + (FN - FS) / dy - src)

    def leap_frog(self, q0):

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        FE = self.hyperbolicFlux(self.edgeE, 1) - self.diffusiveFlux(self.edgeE, self.height.edgeE, self.slip_length.field, 1)
        FN = self.hyperbolicFlux(self.edgeN, 2) - self.diffusiveFlux(self.edgeN, self.height.edgeN, self.slip_length.field, 2)
        FW = self.hyperbolicFlux(self.edgeW, 1) - self.diffusiveFlux(self.edgeW, self.height.edgeW, self.slip_length.field, 1)
        FS = self.hyperbolicFlux(self.edgeS, 2) - self.diffusiveFlux(self.edgeS, self.height.edgeS, self.slip_length.field, 2)
        src = self.get_source(self.field, self.height.field, self.slip_length.field)

        self.field = q0 - self.dt * ((FE - FW) / dx + (FN - FS) / dy - src)

    def cubicInterpolation(self, q, ax):

        # weights proposed by Bell et al., PRE 76 (2007)
        # a1 = (np.sqrt(7) + 1) / 4
        # a2 = (np.sqrt(7) - 1) / 4

        # weights from piecewise parabolic method (PPM)
        # Collela and Woodward, J. Comp. Phys. 54 (1984)
        a1 = 7 / 12
        a2 = 1 / 12

        Q = a1 * (q + np.roll(q, -1, axis=ax)) - a2 * (np.roll(q, 1, axis=ax) + np.roll(q, -2, axis=ax))

        return Q

    def hyperbolicTVD(self):

        Qx = self.cubicInterpolation(self.field, 1)
        Qy = self.cubicInterpolation(self.field, 2)

        Fx = self.hyperbolicFlux(Qx, 1)
        Fy = self.hyperbolicFlux(Qy, 2)

        flux_x = Fx - np.roll(Fx, 1, axis=1)
        flux_y = Fy - np.roll(Fy, 1, axis=2)

        return flux_x, flux_y

    def hyperbolicCD(self, q, p, dt, ax):

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        F = self.hyperbolicFlux(q.field, ax)

        flux = np.roll(F, -1, axis=ax) - np.roll(F, 1, axis=ax)

        if ax == 1:
            dx = self.disc["dx"]
            return dt / (2 * dx) * flux

        elif ax == 2:
            dy = self.disc["dy"]
            return dt / (2 * dy) * flux

    def diffusiveCD(self):

        Dx = self.diffusiveFlux(self.field, self.height.field, self.slip_length.field, 1)
        Dy = self.diffusiveFlux(self.field, self.height.field, self.slip_length.field, 2)

        flux_x = np.roll(Dx, -1, axis=1) - np.roll(Dx, 1, axis=1)
        flux_y = np.roll(Dy, -1, axis=2) - np.roll(Dy, 1, axis=2)

        return flux_x, flux_y



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import os
import time
import netCDF4
import numpy as np

from hans.material import Material
from hans.geometry import GapHeight


class DatasetSelector:

    def __init__(self, path, mode="select", fname=[], prefix=""):

        self._ds = self.get_files(path, prefix=prefix, mode=mode, fname=fname)

    def get_files(self, path, prefix="", mode="select", fname=[]):
        """
        Select netCDF data files for plotting.

        Parameters
        ----------
        path : str
            relative path below which is searched for files
        prefix : str
            filter files that start with prefix, default=""
        mode : str
            can be one of the following, default="select"
            - select: manually select files
            - single: manually select a single file
            - all: select all files found below path with prefix and suffix
            - name: select files by name through 'fname' keyword argument
        fname : list
            list of file names for mode=name, relative to 'path', defaul="[]"
        Returns
        ----------
        out : dict
            dictionary where keys are filenames and values are corresponding datasets.
            Datasets currently only implemented for suffices "nc" (netCDF4.Dataset) and "dat" (numpy.ndarray).
            Else, values are None.
        """

        assert mode in ["single", "select", "all", "name"], "mode must be 'single', select, 'all' or 'name'"

        fileList = []

        for root, dirs, files in os.walk(path, topdown=False):
            for name in files:
                if name.startswith(prefix) and name.endswith(".nc"):
                    fileList.append(os.path.join(root, name))

        fileList = sorted(fileList)

        if mode != "name":
            print("Available files:")
            for i, file in enumerate(fileList):
                date = time.strftime('%d/%m/%Y %H:%M', time.localtime(os.path.getmtime(file)))
                print(f"{i:3d}: {file:<50} {date}")

        if mode == "single":
            mask = [int(input("Enter file key: "))]
        elif mode == "all":
            mask = list(range(len(fileList)))
        elif mode == "select":
            inp = input("Enter file keys (space separated or range [start]-[end] or combination of both): ")

            mask = [int(i) for i in inp.split() if len(i.split("-")) < 2]
            mask_range = [i for i in inp.split() if len(i.split("-")) == 2]

            for j in mask_range:
                mask += list(range(int(j.split("-")[0]), int(j.split("-")[1]) + 1))
        elif mode == "name":
            fileList = [os.path.join(path, f) for f in fname]
            mask = np.arange(len(fileList))

        out = {f: netCDF4.Dataset(f) for i, f in enumerate(fileList) if i in mask}

        return out

    def get_filenames(self):
        return list(self._ds.keys())

    def get_NetCDF_Datasets(self):
        return list(self._ds.values())

    def get_metadata(self):
        category_prefix = ["options", "disc", "bc", "geometry", "roughness", "ic", "numerics", "material", "surface"]
        out = []
        for f in self._ds.values():
            metadict = {}
            for cat in category_prefix:
                catdict = self._generate_input_dict(f, cat)
                if catdict is not None:
                    metadict[cat] = catdict

            metadict["info"] = {}
            for info in ["restarts", "version"]:
                metadict["info"].update({info: f.getncattr(info)})

            out.append(metadict)

        return out

    def _generate_input_dict(self, data, category):
        out = {(k.split("_")[-1]): (v if v != "None" else None) for k, v in dict(data.__dict__).items() if k.startswith(category)}

        if len(out) > 0:
            return out
        else:
            return None

    def _get_grid(self, data):

        Nx = int(data.disc_Nx)
        Ny = int(data.disc_Ny)
        Lx = float(data.disc_Lx)
        Ly = float(data.disc_Ly)

        return Nx, Ny, Lx, Ly

    def get_centerline(self, key=None, index=-1, dir='x'):

        out = []
        keys = ["rho", "p", "jx", "jy"]

        for data in self._ds.values():

            Nx, Ny, Lx, Ly = self._get_grid(data)

            if dir == "x":
                xdata = (np.arange(Nx) + 0.5) * Lx / Nx
            elif dir == "y":
                xdata = (np.arange(Ny) + 0.5) * Ly / Ny

            if key is None:
                ydata = {}
                for k in keys:

                    if k == "p":
                        rho = np.array(data.variables["rho"][index])
                        material = self._generate_input_dict(data, "material")
                        frame = Material(material).eos_pressure(rho)
                    else:
                        frame = np.array(data.variables[k][index])

                    if dir == "x":
                        ydata[k] = frame[:, Ny // 2]
                    elif dir == "y":
                        ydata[k] = frame[Nx // 2, :]

            else:
                assert key in keys

                if key == "p":
                    rho = np.array(data.variables["rho"][index])
                    material = self._generate_input_dict(data, "material")
                    frame = Material(material).eos_pressure(rho)
                else:
                    frame = np.array(data.variables[key][index])
                if dir == "x":
                    ydata = frame[:, Ny // 2]
                elif dir == "y":
                    ydata = frame[Nx // 2, :]

            out.append((xdata, ydata))

        return out

    def get_centerlines(self, key=None, freq=1, dir='x'):

        out = []
        keys = ["rho", "p", "jx", "jy"]

        for data in self._ds.values():

            Nx, Ny, Lx, Ly = self._get_grid(data)

            if dir == "x":
                xdata = (np.arange(Nx) + 0.5) * Lx / Nx
            elif dir == "y":
                xdata = (np.arange(Ny) + 0.5) * Ly / Ny

            time = np.array(data.variables["time"][::freq])

            if key is None:
                ydata = {}
                for k in keys:

                    if k == "p":
                        rho = np.array(data.variables["rho"][::freq])
                        material = self._generate_input_dict(data, "material")
                        frames = Material(material).eos_pressure(rho)
                    else:
                        frames = np.array(data.variables[k][::freq])

                    if dir == "x":
                        ydata[k] = frames[:, :, Ny // 2]
                    elif dir == "y":
                        ydata[k] = frames[:, Nx // 2, :]

            else:
                assert key in keys

                if key == "p":
                    rho = np.array(data.variables["rho"][::freq])
                    material = self._generate_input_dict(data, "material")
                    frames = Material(material).eos_pressure(rho)
                else:
                    frames = np.array(data.variables[key][::freq])

                if dir == "x":
                    ydata = frames[:, :, Ny // 2]
                elif dir == "y":
                    ydata = frames[:, Nx // 2, :]

            out.append((time, xdata, ydata))

        return out

    def get_scalar(self, key=None, freq=1):

        out = []

        keys = ["mass", "eps", "vSound", "vmax", "dt", "ekin"]

        for data in self._ds.values():

            time = np.array(data.variables['time'])[::freq]

            if key is None:
                ydata = {}
                for k in keys:
                    try:
                        ydata[k] = np.array(data.variables[k])[::freq]
                    except KeyError:
                        print(f"Scalar variable {k} not existing.")
                        pass

            else:
                assert key in keys
                time = np.array(data.variables['time'])[::freq]
                ydata = np.array(data.variables[key])[::freq]

            out.append((time, ydata))

        return out

    def get_field(self, key=None, index=-1):

        out = []

        keys = ["rho", "p", "jx", "jy"]

        for data in self._ds.values():

            if key is None:
                zdata = {}
                for k in keys:
                    if k == "p":
                        rho = np.array(data.variables["rho"][index])
                        material = self._generate_input_dict(data, "material")
                        zdata[k] = Material(material).eos_pressure(rho)
                    else:
                        zdata[k] = np.array(data.variables[k][index])

            else:
                assert key in keys
                if key == "p":
                    rho = np.array(data.variables["rho"][index])
                    material = self._generate_input_dict(data, "material")
                    zdata = Material(material).eos_pressure(rho)
                else:
                    zdata = np.array(data.variables[key][index])

            out.append(zdata)

        return out

    def get_height(self):

        out = []

        for data in self._ds.values():
            disc = self._generate_input_dict(data, "disc")
            disc['pX'] = True
            disc['pY'] = True
            disc['nghost'] = 1

            geometry = self._generate_input_dict(data, "geometry")
            roughness = self._generate_input_dict(data, "roughness")
            frame = GapHeight(disc, geometry, roughness)

            out.append(frame.inner[0])

        return out

    def get_centerline_height(self, dir='x'):

        out = []

        for data in self._ds.values():

            Nx, Ny, Lx, Ly = self._get_grid(data)

            if dir == "x":
                xdata = (np.arange(Nx) + 0.5) * Lx / Nx
            elif dir == "y":
                xdata = (np.arange(Ny) + 0.5) * Ly / Ny

            disc = self._generate_input_dict(data, "disc")
            disc['pX'] = True
            disc['pY'] = True
            disc['nghost'] = 1

            geometry = self._generate_input_dict(data, "geometry")
            roughness = self._generate_input_dict(data, "roughness")
            frame = GapHeight(disc, geometry, roughness)

            if dir == "x":
                ydata = frame.centerline_x[0]
            elif dir == "y":
                ydata = frame.centerline_y[0]

            out.append((xdata, ydata))

        return out

    def get_fields(self, key=None, freq=1):

        out = []

        keys = ["rho", "p", "jx", "jy"]

        for data in self._ds.values():

            time = np.array(data.variables["time"][::freq])

            if key is None:
                zdata = {}
                for k in keys:
                    if k == "p":
                        rho = np.array(data.variables["rho"][::freq])
                        material = self._generate_input_dict(data, "material")
                        zdata[k] = Material(material).eos_pressure(rho)
                    else:
                        zdata[k] = np.array(data.variables[k][::freq])
            else:
                assert key in keys

                if key == "p":
                    rho = np.array(data.variables["rho"][::freq])
                    material = self._generate_input_dict(data, "material")
                    zdata = Material(material).eos_pressure(rho)
                else:
                    zdata = np.array(data.variables[key][::freq])

            out.append((time, zdata))

        return out


def adaptiveLimits(ax):

    def offset(x, y): return 0.05 * (x - y) if (x - y) != 0 else 1.

    try:
        axes = ax.flat
    except AttributeError:
        axes = [ax]

    for a in axes:

        y_min = np.amin(a.lines[0].get_ydata())
        y_max = np.amax(a.lines[0].get_ydata())

        a.set_ylim(y_min - offset(y_max, y_min), y_max + offset(y_max, y_min))

    return ax



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np
from mpi4py import MPI

from hans.field import ScalarField, VectorField
from hans.tools import abort


class GapHeight(VectorField):

    def __init__(self, disc, geometry, roughness=None):

        super().__init__(disc)

        self.geometry = geometry
        self.disc = disc
        self.roughness = roughness

        self.set_profile()

        if self.roughness is not None:
            if "file" in self.roughness.keys():
                self.add_roughness_from_file()
            else:
                self.add_roughness()

        self.set_gradients()

    @property
    def max_height(self):
        local_max = np.amax(self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_max, recvbuf, op=MPI.MAX)

        return recvbuf[0]

    @property
    def min_height(self):
        local_max = np.amin(self.inner[0])
        recvbuf = np.empty(1, dtype=float)
        self.comm.Allreduce(local_max, recvbuf, op=MPI.MIN)

        return recvbuf[0]

    def set_profile(self):

        Lx = self.disc["Lx"]
        Ly = self.disc["Ly"]
        Nx = self.disc["Nx"]
        Ny = self.disc["Ny"]
        dx = self.disc["dx"]
        dy = self.disc["dy"]

        idxx, idyy = self.id_grid

        ng = self.disc["nghost"]

        xx = idxx * (Lx + 2 * ng * dx) / (Nx + 2 * ng) + dx / 2
        yy = idyy * (Ly + 2 * ng * dy) / (Ny + 2 * ng) + dy / 2

        if self.geometry["type"] in ["journal", "journal_x"]:
            CR = self.geometry['CR']
            eps = self.geometry['eps']

            Rb = Lx / (2 * np.pi)
            c = CR * Rb
            e = eps * c
            self.field[0] = c + e * np.cos(xx / Rb)

        elif self.geometry["type"] == "journal_y":
            CR = self.geometry['CR']
            eps = self.geometry['eps']

            Rb = Ly / (2 * np.pi)
            c = CR * Rb
            e = eps * c
            self.field[0] = c + e * np.cos(yy / Rb)

        elif self.geometry["type"] == "parabolic":
            hmin = self.geometry['hmin']
            hmax = self.geometry['hmax']
            self.field[0] = 4 * (hmax - hmin) / Lx**2 * (xx - Lx / 2)**2 + hmin

        elif self.geometry["type"] == "twin_parabolic":
            hmin = self.geometry['hmin']
            hmax = self.geometry['hmax']

            right = np.greater(xx, Lx / 2)
            self.field[0] = 16 * (hmax - hmin) / Lx**2 * (xx - Lx / 4)**2 + hmin
            self.field[0][right] = 16 * (hmax - hmin) / Lx**2 * (xx[right] - 3 * Lx / 4)**2 + hmin

        elif self.geometry["type"] in ["inclined", "inclined_x"]:
            h1 = self.geometry['h1']
            h2 = self.geometry['h2']
            self.field[0] = h1 + (h2 - h1) / (Lx) * xx

        elif self.geometry["type"] == "inclined_y":
            h1 = self.geometry['h1']
            h2 = self.geometry['h2']
            self.field[0] = h1 + (h2 - h1) / (Ly) * yy

        elif self.geometry["type"] == "inclined_pocket":
            h1 = self.geometry['h1']
            h2 = self.geometry['h2']
            hp = self.geometry['hp']
            cp = self.geometry['c']
            lp = self.geometry['l']
            wp = self.geometry['w']
            tw = self.geometry['t']

            self.field[0] = h1 + (h2 - h1) / (Lx) * xx

            xmask = np.logical_and(xx > cp, xx <= cp + lp)
            ymask = np.logical_and(yy > (Ly - wp) / 2., yy <= (Ly + wp) / 2.)
            xymask = np.logical_and(xmask, ymask)

            xt1_mask = np.logical_and(xx > cp - tw, xx <= cp)
            xt2_mask = np.logical_and(xx > cp + lp, xx <= cp + lp + tw)

            self.field[0, xt1_mask] += (xx[xt1_mask] - cp + tw) * hp / tw
            self.field[0, xt2_mask] += -(xx[xt2_mask] - cp - lp - tw) * hp / tw

            self.field[0, xymask] += hp

        elif self.geometry["type"] == "half_sine":
            h0 = self.geometry['h0']
            amp = self.geometry['amp']
            num = self.geometry['num']

            self.field[0] = h0 - amp * np.sin(- 4 * np.pi * (xx - Lx / 2) * num / Lx)
            mask = np.greater(xx, Lx / 2)
            self.field[0][mask] = h0

        elif self.geometry["type"] == "half_sine_squared":
            h0 = self.geometry['h0']
            amp = self.geometry['amp']
            num = self.geometry['num']

            self.field[0] = h0 + amp * np.sin(- 4 * np.pi * (xx - Lx / 2) * num / Lx)**2
            mask = np.greater(xx, Lx / 2)
            self.field[0][mask] = h0

    def add_roughness_from_file(self):

        Nx = self.disc["Nx"]
        Ny = self.disc["Ny"]

        # periodic cartesian communicator
        pcomm = self.get_2d_cart_comm(MPI.COMM_WORLD, (Nx, Ny),
                                      periods=(True, True))

        topo = np.load(self.roughness["file"])

        if self.min_height < abs(np.amin(topo)):
            if pcomm.Get_rank() == 0:
                print("Rough topography not compatible with geometry. Abort!")
                abort()

        if topo.shape != (Nx, Ny):
            if pcomm.Get_rank() == 0:
                print(f"Shape mismatch. Expected ({Nx}, {Ny}), but imported topography has {topo.shape}. Abort!")
                abort()

        topo_periodic = self.roughness_pbc(topo, pcomm)

        self.field[0] += topo_periodic

    def add_roughness(self):
        """
        Add periodic, self-affine random roughness profile on top of analytical profiles.
        """

        Nx = self.disc["Nx"]
        Ny = self.disc["Ny"]
        Lx = self.disc["Lx"]
        Ly = self.disc["Ly"]

        # periodic cartesian communicator
        pcomm = self.get_2d_cart_comm(MPI.COMM_WORLD, (Nx, Ny),
                                      periods=(True, True))

        topo = np.ones((Nx, Ny), dtype=np.float64) * 1e8
        repeat = 0
        repeat_limit = 100

        while abs(np.amin(topo)) > self.min_height:
            if pcomm.Get_rank() == 0:
                if repeat == 0:
                    print("Generating random roughness ...")
                    seed = self.roughness["seed"]
                else:
                    seed = None

                np.random.seed(seed)

                hurst = self.roughness["Hurst"]
                rolloff = self.roughness["rolloff"]
                rms_height = self.roughness["rmsHeight"]
                rms_slope = self.roughness["rmsSlope"]
                short_cutoff = self.roughness["shortCutoff"]
                long_cutoff = self.roughness["longCutoff"]

                topo = fourier_synthesis((Nx, Ny), (Lx, Ly), hurst,
                                         rms_height=rms_height,
                                         rms_slope=rms_slope,
                                         short_cutoff=short_cutoff,
                                         long_cutoff=long_cutoff,
                                         rolloff=rolloff)
                repeat += 1

                if repeat == repeat_limit:
                    print("Could not generate valid topography. Abort!")
                    abort()

            pcomm.Bcast(topo)

        else:
            if pcomm.Get_rank() == 0:
                if repeat == 1:
                    print("Roughness generation successful!")
                else:
                    print(f"Roughness generation successful (tried {repeat} seeds)!")
                print(60 * "-")

        topo_periodic = self.roughness_pbc(topo, pcomm)

        # add roughness
        self.field[0] += topo_periodic

    def roughness_pbc(self, topo, pcomm):

        ng = self.disc["nghost"]
        ngt = 2 * ng
        topo_periodic = np.empty_like(self.field[0])
        topo_periodic[ng:-ng, ng:-ng] = topo[self.without_ghost]

        ls, ld = pcomm.Shift(0, -1)
        rs, rd = pcomm.Shift(0, 1)
        bs, bd = pcomm.Shift(1, -1)
        ts, td = pcomm.Shift(1, 1)

        # Send to left, receive from right
        recvbuf = np.ascontiguousarray(topo_periodic[-ng:, :])
        pcomm.Sendrecv(np.ascontiguousarray(topo_periodic[ng:ngt, :]), ld, recvbuf=recvbuf, source=ls)
        # fill ghost buffer
        topo_periodic[-ng:, :] = recvbuf

        # Send to right, receive from left
        recvbuf = np.ascontiguousarray(topo_periodic[:ng, :])
        pcomm.Sendrecv(np.ascontiguousarray(topo_periodic[-ngt:-ng, :]), rd, recvbuf=recvbuf, source=rs)
        topo_periodic[:ng, :] = recvbuf

        # Send to bottom, receive from top
        recvbuf = np.ascontiguousarray(topo_periodic[:, -ng:])
        pcomm.Sendrecv(np.ascontiguousarray(topo_periodic[:, ng:ngt]), bd, recvbuf=recvbuf, source=bs)
        # fill ghost buffer
        topo_periodic[:, -ng:] = recvbuf

        # Send to top, receive from bottom
        recvbuf = np.ascontiguousarray(topo_periodic[:, :ng])
        pcomm.Sendrecv(np.ascontiguousarray(topo_periodic[:, -ngt:-ng]), td, recvbuf=recvbuf, source=ts)
        # fill ghost buffer
        topo_periodic[:, :ng] = recvbuf

        return topo_periodic

    def set_gradients(self):
        "gradients for a scalar field (1st entry), stored in 2nd (dx) and 3rd (dy) entry of vectorField"

        dx = self.disc["dx"]
        dy = self.disc["dy"]

        self.field[1:] = np.gradient(self.field[0], dx, dy, edge_order=2)


def fourier_synthesis(shape, size, Hurst, rms_height=None, rms_slope=None,
                      short_cutoff=None, long_cutoff=None, rolloff=1.0):
    """
    Create a self-affine, randomly rough surface using a Fourier filtering
    algorithm. Adapted from https://github.com/ContactEngineering/SurfaceTopography.
    The algorithm is described in:
    Ramisetti et al., J. Phys.: Condens. Matter 23, 215004 (2011);
    Jacobs, Junge, Pastewka, Surf. Topgogr.: Metrol. Prop. 5, 013001 (2017)

    Parameters
    ----------
    shape : tuple of ints
        Number of grid cells in x and y direction
    size : tuple of floats
        Physical size in x and y direction
    Hurst : float
        Hurst exponent.
    rms_height : float
        Root mean-squared height.
    rms_slope : float
        Root mean-squared slope.
    short_cutoff : float
        Short-wavelength cutoff.
    long_cutoff : float
        Long-wavelength cutoff.
    rolloff : float
        Value for the power-spectral density (PSD) below the long-wavelength
        cutoff. This multiplies the value at the cutoff, i.e. unit will give a
        PSD that is flat below the cutoff, zero will give a PSD that is
        vanishes below cutoff. (Default: 1.0)

    Returns
    -------
    np.array
        Rough topography.
    """

    Nx, Ny = shape
    Lx, Ly = size

    transpose = False

    if Nx == 1 and Ny > 1:
        grid_pts = (Ny,)
        physical_sizes = (Ly,)
    elif Ny == 1 and Nx > 1:
        grid_pts = (Nx,)
        physical_sizes = (Lx,)
        transpose = True
    else:
        grid_pts = (Nx, Ny)
        physical_sizes = (Lx, Ly)

    if short_cutoff is not None:
        q_max = 2 * np.pi / short_cutoff
    else:
        q_max = np.pi * np.min(np.asarray(grid_pts) / np.asarray(physical_sizes))

    if long_cutoff is not None:
        q_min = 2 * np.pi / long_cutoff
    else:
        q_min = None

    fac = self_affine_prefactor(grid_pts, physical_sizes, Hurst, rms_height=rms_height,
                                rms_slope=rms_slope, short_cutoff=short_cutoff,
                                long_cutoff=long_cutoff)

    if len(grid_pts) == 2:
        nx, ny = grid_pts
        sx, sy = physical_sizes
        kny = ny // 2 + 1
        kshape = (nx, kny)
    else:
        nx = 1
        ny, = grid_pts
        sx = 1
        sy, = physical_sizes
        kny = ny // 2 + 1
        kshape = (kny,)

    # Initialize arrays
    rarr = np.empty((nx, ny), dtype=np.float64)
    karr = np.empty(kshape, dtype=np.complex128)

    # Creating Fourier representation
    qy = 2*np.pi*np.arange(kny)/sy
    for x in range(nx):
        if x > nx//2:
            qx = 2 * np.pi * (nx - x) / sx
        else:
            qx = 2 * np.pi * x / sx
        q_sq = qx**2 + qy**2
        if x == 0:
            q_sq[0] = 1.
        phase = np.exp(2 * np.pi * np.random.rand(kny) * 1j)
        ran = fac * phase * np.random.normal(size=kny)  # amplitudes are normal distributed

        if len(kshape) == 2:
            karr[x, :] = ran * q_sq ** (-(1 + Hurst) / 2)
            karr[x, q_sq > q_max ** 2] = 0.
        else:
            karr[:] = ran * q_sq ** (-(0.5 + Hurst) / 2)
            karr[q_sq > q_max ** 2] = 0.

        if q_min is not None:
            mask = q_sq < q_min**2
            if len(kshape) == 2:
                karr[x, mask] = rolloff * ran[mask] * q_min ** (-(1 + Hurst))
            else:
                karr[mask] = rolloff * ran[mask] * q_min ** (-(0.5 + Hurst))

    # Inverse FFT
    if len(kshape) == 2:
        for iy in [0, -1] if ny % 2 == 0 else [0]:
            # Enforce symmetry
            if nx % 2 == 0:
                karr[0, iy] = np.real(karr[0, iy])
                karr[nx // 2, iy] = np.real(karr[nx // 2, iy])
                karr[1:nx // 2, iy] = karr[-1:nx // 2:-1, iy].conj()
            else:
                karr[0, iy] = np.real(karr[0, iy])
                karr[1:nx // 2 + 1, iy] = karr[-1:nx // 2:-1, iy].conj()
        _irfft2(karr, rarr)
    else:
        karr[0] = np.real(karr[0])
        rarr[:] = np.fft.irfft(karr, n=ny)

    # Shift to zero mean
    rarr -= np.mean(rarr)

    if transpose:
        return rarr.T
    else:
        return rarr


def self_affine_prefactor(grid_pts, physical_sizes, Hurst, rms_height=None,
                          rms_slope=None, short_cutoff=None, long_cutoff=None):
    r"""
    Compute prefactor :math:`C_0` for the power-spectrum density of an ideal
    self-affine topography given by

    .. math ::

        C(q) = C_0 q^{-2-2H}

    for two-dimensional topography maps and

    .. math ::

        C(q) = C_0 q^{-1-2H}

    for one-dimensional line scans. Here :math:`H` is the Hurst exponent.

    Note:
    In the 2D case:

    .. math ::

        h^2_{rms} = \frac{1}{2 \pi} \int_{0}^{\infty} q C^{iso}(q) dq

    whereas in the 1D case:

    .. math ::

        h^2_{rms} = \frac{1}{\pi} \int_{0}^{\infty} C^{1D}(q) dq

    See Equations (1) and (4) in [1].


    Parameters
    ----------
    grid_pts : array_like
        Resolution of the topography map or the line scan.
    physical_sizes : array_like
        Physical physical_sizes of the topography map or the line scan.
    Hurst : float
        Hurst exponent.
    rms_height : float
        Root mean-squared height.
    rms_slope : float
        Root mean-squared slope of the topography map or the line scan.
    short_cutoff : float
        Short-wavelength cutoff.
    long_cutoff : float
        Long-wavelength cutoff.

    Returns
    -------
    prefactor : float
        Prefactor :math:`\sqrt{C_0}`

    References
    -----------
    [1]: Jacobs, Junge, Pastewka,
    Surf. Topgogr.: Metrol. Prop. 5, 013001 (2017)

    """

    grid_pts = np.asarray(grid_pts)
    physical_sizes = np.asarray(physical_sizes)

    if short_cutoff is not None:
        q_max = 2 * np.pi / short_cutoff
    else:
        q_max = np.pi * np.min(grid_pts / physical_sizes)

    if long_cutoff is not None:
        q_min = 2 * np.pi / long_cutoff
    else:
        q_min = 2 * np.pi * np.max(1 / physical_sizes)

    area = np.prod(physical_sizes)

    if rms_height is not None:
        # Assuming no rolloff region
        fac = 2 * rms_height / np.sqrt(q_min**(-2 * Hurst) -
                                       q_max**(-2 * Hurst)) * np.sqrt(Hurst * np.pi)
    elif rms_slope is not None:
        fac = 2 * rms_slope / np.sqrt(q_max**(2 - 2 * Hurst) -
                                      q_min**(2 - 2 * Hurst)) * np.sqrt((1 - Hurst) * np.pi)
    else:
        # This is already caught in sanity checks
        raise ValueError('Neither rms height nor rms slope is defined!')

    return fac * np.prod(grid_pts) / np.sqrt(area)


def _irfft2(karr, rarr):
    """
    Inverse 2d real-to-complex FFT.

    Parameters
    ----------
    karr : array_like
        Fourier-space representation
    rarr : array_like
        Real-space representation
    """
    nrows, ncolumns = karr.shape
    for i in range(ncolumns):
        karr[:, i] = np.fft.ifft(karr[:, i])
    for i in range(nrows):
        if rarr.shape[1] % 2 == 0:
            rarr[i, :] = np.fft.irfft(karr[i, :])
        else:
            rarr[i, :] = np.fft.irfft(karr[i, :], n=rarr.shape[1])


class SlipLength(ScalarField):

    def __init__(self, disc, surface):

        super().__init__(disc)

        if surface is not None:

            self.surface = surface
            self.disc = disc

            self.fill()

    def fill(self):

        Lx = self.disc["Lx"]
        Ly = self.disc["Ly"]
        Nx = self.disc["Nx"]
        Ny = self.disc["Ny"]
        dx = self.disc["dx"]
        dy = self.disc["dy"]

        idxx, idyy = self.id_grid

        ng = self.disc["nghost"]

        xx = idxx * (Lx + 2 * ng * dx) / (Nx + 2 * ng) + dx / 2
        yy = idyy * (Ly + 2 * ng * dy) / (Ny + 2 * ng) + dy / 2

        if self.surface["type"] in ["stripes", "stripes_x"]:
            num = self.surface["num"]
            sign = self.surface["sign"]
            sin = sign * np.sin(2 * np.pi * xx * num / Lx)
            mask = np.greater(sin, 0)

        elif self.surface["type"] in ["stripes_y"]:
            num = self.surface["num"]
            sign = self.surface["sign"]
            sin = sign * np.sin(2 * np.pi * yy * num / Ly)
            mask = np.greater(sin, 0)

        elif self.surface["type"] == "checkerboard":
            right_x = np.greater(xx, Lx / 2)
            center_y = np.logical_and(np.greater(yy, Ly / 4), np.less(yy, 3 * Ly / 4))
            mask_1 = np.logical_and(right_x, center_y)
            mask_2 = np.logical_and(np.logical_not(right_x), np.logical_not(center_y))
            mask = np.logical_or(mask_1, mask_2)

        elif self.surface["type"] == "circle":
            center = (3 * Lx / 4, Ly / 2)
            radius = Lx / 4
            mask = np.less((xx - center[0])**2 + (yy - center[1])**2, radius**2)

        elif self.surface["type"] == "circle2":
            center_1 = (3 * Lx / 4, Ly/2)
            center_2 = (Lx / 4, Ly)
            center_3 = (Lx / 4, 0)
            radius = Lx / 4
            mask_1 = np.less((xx - center_1[0])**2 + (yy - center_1[1])**2, radius**2)
            mask_2 = np.less((xx - center_2[0])**2 + (yy - center_2[1])**2, radius**2)
            mask_3 = np.less((xx - center_3[0])**2 + (yy - center_3[1])**2, radius**2)

            mask = np.logical_or(np.logical_or(mask_1, mask_2), mask_3)

        elif self.surface["type"] == "square":
            right_x = np.greater(xx, Lx / 2)
            center_y = np.logical_and(np.greater(yy, Ly / 4), np.less(yy, 3 * Ly / 4))
            mask = np.logical_and(right_x, center_y)

        elif self.surface["type"] == "full":
            mask = None

        ls = self.surface["lslip"]
        self.field[0, mask] = ls



#
#


import os
import sys
from collections import defaultdict
from datetime import datetime
from subprocess import Popen, PIPE

root = os.path.dirname(sys.argv[0])


def read_authors(fn):
    return {email.strip('<>'): name for name, email in
            [line.rsplit(maxsplit=1) for line in open(fn, 'r')]}


def parse_git_log(log, authors):
    committers = defaultdict(set)
    author = None
    date = None
    for line in log.decode('utf-8').split('\n'):
        if line.startswith('commit'):
            if date is not None and author is not None:
                committers[author].add(date.year)
        elif line.startswith('Author:'):
            email = line.rsplit('<', maxsplit=1)[1][:-1]
        elif line.startswith('Date:'):
            date = datetime.strptime(line[5:].rsplit(maxsplit=1)[0].strip(),
                                     '%a %b %d %H:%M:%S %Y')
            try:
                author = authors[email]
            except KeyError:
                author = email
        elif 'copyright' in line.lower() or 'license' in line.lower():
            date = None
    if date is not None:
        committers[author].add(date.year)
    return committers


def pretty_years(years):
    years = sorted(years)
    prev_year = prev_out = years[0]
    s = '{}'.format(prev_year)
    for year in years[1:]:
        if year - prev_year > 1:
            if year - prev_out > 1:
                if prev_year == prev_out:
                    s = '{}, {}'.format(s, year)
                else:
                    s = '{}-{}, {}'.format(s, prev_year, year)
            else:
                s = '{}, {}'.format(s, prev_year)
            prev_out = year
        prev_year = year
    if prev_year - prev_out == 1:
        s = '{}-{}'.format(s, prev_year)
    elif prev_year - prev_out > 1:
        s = '{}, {}'.format(s, prev_year)
    return s


authors = read_authors('{}/../AUTHORS'.format(root))

process = Popen(['git', 'log', '--follow', sys.argv[1]], stdout=PIPE,
                stderr=PIPE)
stdout, stderr = process.communicate()
committers = parse_git_log(stdout, authors)

prefix = 'Copyright'
for name, years in committers.items():
    print('{} {} {}'.format(prefix, pretty_years(years), name))
    prefix = ' ' * len(prefix)
print()



#
# Copyright 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from argparse import ArgumentParser

from hans.geometry import fourier_synthesis


def get_parser():
    parser = ArgumentParser()
    parser.add_argument('filename', metavar="FILENAME")
    parser.add_argument('-Nx',
                        dest="Nx",
                        default=128,
                        type=int,
                        help="Nx (default: 128)")
    parser.add_argument('-Ny',
                        dest="Ny",
                        default=128,
                        type=int,
                        help="Ny (default: 128)")
    parser.add_argument('-Lx',
                        dest="Lx",
                        default=1.,
                        type=float,
                        help="Lx (default: 1.)")
    parser.add_argument('-Ly',
                        dest="Ly",
                        default=1.,
                        type=float,
                        help="Ly (default: 1.)")
    parser.add_argument('-H', '--Hurst',
                        dest="hurst",
                        default=0.8,
                        type=float,
                        help="Hurst exponent (default: 0.8)")
    parser.add_argument('--rms-height',
                        dest="rms_height",
                        default=None,
                        type=float,
                        help="Root-mean squared height (default: None)")
    parser.add_argument('--rms-slope',
                        dest="rms_slope",
                        default=None,
                        type=float,
                        help="Root-mean squared slope (default: None)")
    parser.add_argument('--short-cutoff',
                        dest="short_cutoff",
                        default=None,
                        type=float,
                        help="Short wavelength cutoff (default: None)")
    parser.add_argument('--long-cutoff',
                        dest="long_cutoff",
                        default=None,
                        type=float,
                        help="Long wavelength cutoff (default: None)")
    parser.add_argument('--rolloff',
                        dest="rolloff",
                        default=1.0,
                        type=float,
                        help="Rolloff (default: 1.0)")
    parser.add_argument('--seed',
                        dest="seed",
                        default=None,
                        type=int,
                        help="Random seed (default: None)")
    parser.add_argument('--plot',
                        dest="plot",
                        default=False,
                        action="store_true",
                        help="Show surface plot (default: False)")

    return parser


def plot_surface(arr):
    fig, ax = plt.subplots(subplot_kw={"projection": "3d"}, tight_layout=True)

    x = np.arange(arr.shape[0]) / arr.shape[0]
    y = np.arange(arr.shape[1]) / arr.shape[1]

    xx, yy = np.meshgrid(x, y)

    ax.plot_surface(xx, yy, arr, cmap=cm.viridis, linewidth=0, antialiased=False)

    plt.show()


def main():
    parser = get_parser()
    args = parser.parse_args()

    np.random.seed(args.seed)

    topo = fourier_synthesis((args.Nx, args.Ny),
                             (args.Lx, args.Ly),
                             args.hurst,
                             rms_height=args.rms_height,
                             rms_slope=args.rms_slope,
                             short_cutoff=args.short_cutoff,
                             long_cutoff=args.long_cutoff,
                             rolloff=args.rolloff)

    print(f"Created height profile ranging from {np.amin(topo):.5e} to {np.amax(topo):.5e}.")

    np.save(args.filename, topo)

    if args.plot:
        plot_surface(topo)


if __name__ == "__main__":
    main()



#
# Copyright 2019, 2021 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


from mpi4py import MPI
import numpy as np


class Field:

    def __init__(self, disc, ndim):
        """Base class for all field objects.

        Parameters
        ----------
        disc : dict
            Discretization parameters.
        ndim : int
            Dimensionality.

        Attributes
        ----------
        comm : mpi4py.MPI.Cartcomm
            2D cartesian communicator.
        field : np.array
            Buffer holding the data of the class instance.
        disc : dict
            Discretization parameters.

        """

        self.disc = disc

        # global array size
        Nx = int(disc['Nx'])
        Ny = int(disc['Ny'])

        # read periodic boundary conditions
        periodicX = bool(disc['pX'])
        periodicY = bool(disc['pY'])

        # get cartesian communicator
        self.comm = self.get_2d_cart_comm(MPI.COMM_WORLD, (Nx, Ny),
                                          periods=(periodicX, periodicY))
        rank = self.comm.Get_rank()
        mpix, mpiy = self.comm.Get_coords(rank)
        size_x, size_y = self.comm.dims

        # local array sizes
        nx = Nx // size_x
        ny = Ny // size_y

        # fill right and topmost grid, s.t. local arrays sizes sum up to global size
        if mpix + 1 == size_x:
            nx += Nx % size_x
        if mpiy + 1 == size_y:
            ny += Ny % size_y

        # bounds of FD grid
        # left/bottom
        xleft = mpix * nx
        ybottom = mpiy * ny

        # right/top
        xright = xleft + nx - 1
        ytop = ybottom + ny - 1

        # adjust rightmost/topmost cell
        if mpix + 1 == size_x:
            xleft = Nx - nx
            xright = Nx - 1
        if mpiy + 1 == size_y:
            ybottom = Ny - ny
            ytop = Ny - 1

        ng = disc["nghost"]

        self._idx = np.linspace(xleft - ng, xright + ng, nx + 2 * ng, dtype=int)
        self._idy = np.linspace(ybottom - ng, ytop + ng, ny + 2 * ng, dtype=int)

        self.field = np.zeros(shape=(ndim, nx + 2 * ng, ny + 2 * ng), dtype=np.float64)

    def get_2d_cart_comm(self, comm, shape, periods=(True, True)):
        """
        Creates a two-dimensional cartesian MPI communicator.

        Parameters
        ----------
        comm : mpi4py.MPI.Intracomm
            MPI communicator object
        periods : Iterable
            Directions with periodic boundary conditions (the default is (True, True)).

        Returns
        -------
        cartcomm : mpi4py.MPI.Cartcomm
            2D cartesian communicator
        rank : int
            rank of the MPI process
        size_x : int
            size of MPI grid in x direction
        size_y : int
            size of MPI grid in y direction

        """

        # get size of communicator
        size = comm.Get_size()

        Nx, Ny = shape

        # define optimal MPI grid from number of procs and global grid size
        possible_grids = [(i, size // i) for i in range(1, size + 1) if size % i == 0]
        grid_selector = [abs(i[0] / i[1] - Nx / Ny) for i in possible_grids]
        size_x, size_y = possible_grids[grid_selector.index(min(grid_selector))]

        # create cartesian communicator
        cartcomm = comm.Create_cart(dims=(size_x, size_y), periods=periods)

        return cartcomm

    def get_neighbors(self):
        """
        Get MPI ranks of neighboring cells.

        Returns
        -------
        tuple
            Contains left, right, bottom and top neighbors, src and destination respectively.

        """
        # get MPI ranks of neighboring cells
        left = self.comm.Shift(0, -1)
        right = self.comm.Shift(0, 1)
        bottom = self.comm.Shift(1, -1)
        top = self.comm.Shift(1, 1)

        return (left, right, bottom, top)

    @property
    def inner(self):
        ng = self.disc["nghost"]
        return np.ascontiguousarray(self.field[:, ng:-ng, ng:-ng])

    @property
    def edgeE(self):
        return (self.field + np.roll(self.field, -1, axis=1)) / 2.

    @property
    def edgeN(self):
        return (self.field + np.roll(self.field, -1, axis=2)) / 2.

    @property
    def edgeW(self):
        return (self.field + np.roll(self.field, 1, axis=1)) / 2.

    @property
    def edgeS(self):
        return (self.field + np.roll(self.field, 1, axis=2)) / 2.

    @property
    def verticeNE(self):
        return (self.field +
                np.roll(self.field, -1, axis=1) +
                np.roll(self.field, -1, axis=2) +
                np.roll(self.field, (-1, -1), axis=(1, 2))) / 4.

    @property
    def verticeSW(self):
        return (self.field +
                np.roll(self.field, 1, axis=1) +
                np.roll(self.field, 1, axis=2) +
                np.roll(self.field, (1, 1), axis=(1, 2))) / 4.

    @property
    def id_grid(self):
        return np.meshgrid(self._idx, self._idy, indexing='ij')

    @property
    def without_ghost(self):
        ng = self.disc["nghost"]
        wo_ghost_x = slice(self._idx[ng], self._idx[-ng])
        wo_ghost_y = slice(self._idy[ng], self._idy[-ng])
        return wo_ghost_x, wo_ghost_y

    @property
    def centerline_x(self):
        ng = self.disc["nghost"]
        return self.field[:, ng:-ng, self.disc["Ny"] // 2 + ng]

    @property
    def centerline_y(self):
        ng = self.disc["nghost"]
        return self.field[:, self.disc["Nx"] // 2 + ng, ng:-ng]


class ScalarField(Field):
    """Scalar field derived from Field base class."""

    def __init__(self, disc):
        """Constructor.

        Parameters
        ----------
        disc : dict
            Description of parameter `disc`.

        Returns
        -------
        type
            Description of returned object.

        """
        super().__init__(disc, 1)


class VectorField(Field):
    """Vector field derived from Field base class."""

    def __init__(self, disc):
        """Constructor.

        Parameters
        ----------
        disc : dict
            Discretization parameters.

        """
        super().__init__(disc, 3)


class TensorField(Field):
    """Symmetric tensor field derived from Field base class."""

    def __init__(self, disc):
        """Constructor.

        Parameters
        ----------
        disc : dict
            Discretization parameters.

        """
        super().__init__(disc, 6)



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import numpy as np
import pytest

from hans.input import Input
from hans.plottools import DatasetSelector


@pytest.fixture(scope="session", params=["MC", "RK3", "LW"])
def setup(tmpdir_factory, request):
    config_file = os.path.join("tests", "journal-bearing1D_x_incompressible.yaml")
    tmp_dir = tmpdir_factory.mktemp("tmp")

    myTestProblem = Input(config_file).getProblem()
    myTestProblem.numerics["integrator"] = request.param
    myTestProblem.run(out_dir=tmp_dir)

    ds = DatasetSelector(tmp_dir, mode="name", fname=[tmp_dir.join(os.path.basename(myTestProblem.outpath))])

    data = ds.get_centerline()[0]
    mass = ds.get_scalar(key="mass")[0]

    yield data, mass


def test_pressure(setup):
    p_ref = np.loadtxt(os.path.join("tests", "journal-bearing1D_eps0.7_incompressible.dat"), usecols=(1,))

    data, _ = setup

    xdata, ydata = data
    p = ydata["p"] / 1e6
    p_ref /= 1e6

    np.testing.assert_almost_equal(p, p_ref, decimal=1)


def test_density(setup):
    rho_ref = np.loadtxt(os.path.join("tests", "journal-bearing1D_eps0.7_incompressible.dat"), usecols=(0,))

    data, _ = setup

    xdata, ydata = data
    rho = ydata["rho"]

    np.testing.assert_almost_equal(rho, rho_ref, decimal=1)


def test_massConservation(setup):

    _, data = setup

    time, mass = data

    relDiff = abs(mass[-1] - mass[0]) / mass[0]
    assert relDiff < 0.001
